# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2017 Tomas Vondra (Launchpad id: tomas-v)
# Copyright 2017 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

import duplicity.backend
from duplicity import log
from duplicity.errors import BackendException

import os
import subprocess
import re

class Megav2Backend(duplicity.backend.Backend):
    """ Backend for MEGA.nz cloud storage, only one that works for accounts created since Nov. 2018 """
    """ See https://github.com/megous/megatools/issues/411 for more details                         """
    """                                                                                             """
    """ This MEGA backend resorts to official tools (MEGAcmd) as available at https://mega.nz/cmd   """
    """ MEGAcmd works through a single binary called "mega-cmd", which talks to a backend server    """
    """ "mega-cmd-server", which keeps state (for example, persisting a session). Multiple "mega-*" """
    """ shell wrappers (ie. "mega-ls") exist as the user interface to "mega-cmd" and MEGA API       """
    """ The full MEGAcmd User Guide can be found in the software's GitHub page below :              """
    """ https://github.com/meganz/MEGAcmd/blob/master/UserGuide.md                                  """

    def __init__(self, parsed_url):
        duplicity.backend.Backend.__init__(self, parsed_url)

        # ensure all the necessary megatools binaries exist
        self._check_binary_exists('mega-login')
        self._check_binary_exists('mega-logout')
        self._check_binary_exists('mega-cmd')
        self._check_binary_exists('mega-cmd-server')
        self._check_binary_exists('mega-ls')
        self._check_binary_exists('mega-mkdir')
        self._check_binary_exists('mega-get')
        self._check_binary_exists('mega-put')
        self._check_binary_exists('mega-rm')

        # store some basic info
        self._hostname = parsed_url.hostname

        #### Although "MEGAcmd" does not use a config file for post Nov 2018 accounts, keeping a file with them following earlier MEGA version syntax comes in handy ($HOME/.megav2rc)
        if parsed_url.password is None:
            self._megarc = os.getenv('HOME') + '/.megav2rc'
            try:
                conf_file = open(self._megarc, "r")
            except Exception as e:
                raise BackendException("No password provided in URL, and MEGA configuration file does not exist as '%s'" %(self._megarc,))

            myvars = {}
            for line in conf_file:
                name, var = line.partition("=")[::2]
                myvars[name.strip()] = str(var.strip())
            conf_file.close()
            self._username = myvars["Username"] 
            self._password = myvars["Password"] 

        else:
            self._username = parsed_url.username
            self._password = self.get_password()

        # remote folder ("MEGAcmd" no longer shows "Root/" at the top of the hierarchy)
        self._folder = '/' + parsed_url.path[1:]

        # Only create the remote folder if it doesn't exist yet
        self.mega_login()
        cmd = ['mega-ls', self._folder]
        try:
            self.subprocess_popen(cmd)
        except Exception as e:
            self._makedir(self._folder)

    def _check_binary_exists(self, cmd):
        'checks that a specified command exists in the current path'

        try:
            # ignore the output, we only need the return code
            subprocess.check_output(['which', cmd])
        except Exception as e:
            raise BackendException("command '%s' not found, make sure 'MEGAcmd' tools (https://mega.nz/cmd) are installed" % (cmd,))

    def _makedir(self, path):
        'creates a remote directory (recursively if necessary)'

        self.mega_login()
        cmd = ['mega-mkdir', '-p', path]
        try:
            self.subprocess_popen(cmd)
        except Exception as e:
            error_str = str(e)
            if "Folder already exists" in error_str:
                raise BackendException("Folder '%s' could not be created on MEGA because it already exists. Use another path or remove the folder in MEGA manually" % (path,))
            else:
                raise BackendException("Folder '%s' could not be created, reason : '%s'" % (path,e))

    def _put(self, source_path, remote_filename):
        'uploads file to MEGA (deletes it first, to ensure it does not exist)'

        try:
            self.delete(remote_filename.decode())
        except Exception:
            pass
        self.upload(local_file=source_path.get_canonical().decode(), remote_file=remote_filename.decode())

    def _get(self, remote_filename, local_path):
        'downloads file from MEGA'

        self.download(remote_file=remote_filename.decode(), local_file=local_path.name.decode())

    def _list(self):
        'list files in the backup folder'

        return self.folder_contents(files_only=True)

    def _delete(self, filename):
        'deletes remote '

        self.delete(remote_file=filename)

    def _close(self):
        'Function called when backend is done being used'

        cmd = ['mega-logout']
        self.subprocess_popen(cmd)

    def mega_login(self):
        'helper function to call from each "duplicity" API method for the backend to make sure there is a session to MEGA, or create one'

        # see if there is a session already established with MEGA, try to open one if not, and timeout in "mega-session" if this gets stuck
        try:
            subprocess.check_output('mega-session', timeout=30)
        except subprocess.TimeoutExpired:
            raise BackendException("Timed out while trying to determine if a MEGA session exists")
        except Exception as e:
            cmd = ['mega-login', self._username, self._password]
            try:
                subprocess.check_output(cmd)
            except Exception as e:
                raise BackendException("Could not log in to MEGA, error : '%s'" % (e,))

    def folder_contents(self, files_only=False):
        'lists contents of a folder, optionally ignoring subdirectories'

        #print ("mega-ls -l: %s" % (self._folder,))

        cmd = ['mega-ls', '-l', self._folder]

        self.mega_login()
        files = subprocess.check_output(cmd)
        files = files.decode().split('\n')

        # optionally ignore directories
        if files_only:
            files = [f.split()[5] for f in files if re.search('^-', f)]

        return files

    def download(self, remote_file, local_file):

        #print ("mega-get: %s" % (remote_file,))
        cmd = ['mega-get', self._folder + '/' + remote_file, local_file]
        self.mega_login()
        self.subprocess_popen(cmd)

    def upload(self, local_file, remote_file):

        #print ("mega-put: %s" % (remote_file,))
        cmd = ['mega-put', local_file, self._folder + '/' + remote_file]
        self.mega_login()
        self.subprocess_popen(cmd)

    def delete(self, remote_file):

        #print ("mega-rm: %s" % (remote_file,))
        cmd = ['mega-rm', self._folder + '/' + remote_file]
        self.mega_login()
        self.subprocess_popen(cmd)

duplicity.backend.register_backend('megav2', Megav2Backend)
duplicity.backend.uses_netloc.extend(['megav2'])
